package musr.client.dbchecker;
import javax.swing.*;
import java.io.*;
import java.util.*;

public class DBDiagnostic
{
    private File dbFile;        //File being evaluated
    private Vector message;     //Holds Strings containing descriptions of problems
    private BufferedReader inputStream = null;     //For reading from the .db file.
    public static int READ_AHEAD_BUFFER = 100000;     //Used for max size of buffer for .mark()
    private String[] labelHolder;      //Stores column labels
    private boolean labelsHaveBeenChecked;      //Checks if method checkLabelFormat() has been called
    
    public DBDiagnostic()
    {
        dbFile = null;
        message = new Vector();
        labelHolder = new String[0];
        labelsHaveBeenChecked = false;
    }
    
    public DBDiagnostic(File newDB)
    {
        dbFile = newDB;
        READ_AHEAD_BUFFER = (int)newDB.length();
        message = new Vector();
        labelHolder = new String[0];
        labelsHaveBeenChecked = false;
    }
    
    public void diagnose()              //Just to make it easier to call them both at once.
    {
        if(this.checkFile())
            this.checkContents();
    }
    
    public void setFile(File newDB)     //Could be useful if this was being used in a loop
    {
        dbFile = newDB;
        READ_AHEAD_BUFFER = (int)newDB.length();
    }
    
    public void reset()          //Could be useful if this was being used in a loop
    {
        dbFile = null;
        message = new Vector();
        labelsHaveBeenChecked = false;
    }
    
    public boolean checkFile()             //Checks basics of file's structure and returns true if everyhing is normal.
    {
        if(this.dbFile==null)
        {
            this.message.addElement("No file is currently being checked.");
            return false;
        }
        else
        {
            if(!this.dbFile.exists())
            {    
                this.message.addElement("The selected .db file does not exist.");
                return false;
            }
            if(!this.dbFile.isFile())
            {
                this.message.addElement("The selected .db file is not a file.");
                return false;
            }
            if(!this.dbFile.canRead())
            {
                this.message.addElement("The selected .db file is not readable.");
                return false;
            }
            if(!this.hasDBExtension())
            {
                this.message.addElement("The selected .db file does not appear to be of type .db.");
                return false;
            }
            return true;
        }
    }
    
    private boolean hasDBExtension()        //Checks if file's extension is .db.
    {
        String filename = this.dbFile.getName();
        filename = filename.toLowerCase();
        return(filename.indexOf(".db")>0);
    }
    
    public void checkContents()         //Checks contents of file to ensure they match up with what DBFileReader needs to read it.
    {
        try
        {
            inputStream = new BufferedReader(new FileReader(dbFile));
            inputStream.mark(READ_AHEAD_BUFFER);
            getLabels();
            String currentLine = inputStream.readLine();
            if(currentLine==null)
                this.message.add("This .db file appears to be empty.");
            String command;
            while(currentLine!=null)
            {
                command = getCommand(currentLine);
                if((command.equals("DATA"))||(command.equals("DAT"))||(command.equals("DA"))||(command.equals("D")))
                {
                    inputStream.mark(READ_AHEAD_BUFFER);
                    checkDataFormat(currentLine);
                }
                else if((command.equals("DEFINE"))||(command.equals("DEFIN"))||(command.equals("DEFI"))||(command.equals("DEF")))
                    checkDefineFormat(currentLine);
                else if((command.equals("DELETE"))||(command.equals("DELET"))||(command.equals("DELE"))||(command.equals("DEL")))
                    checkDeleteFormat(currentLine);
                else if((command.equals("GENERATE"))||(command.equals("GENERAT"))||(command.equals("GENERA"))||(command.equals("GENER"))||(command.equals("GENE"))||(command.equals("GEN"))||(command.equals("GE"))||(command.equals("G")))
                    checkGenerateFormat(currentLine);
                else if((!labelsHaveBeenChecked)&&((command.equals("LABELS"))||(command.equals("LABEL"))||(command.equals("LABE"))||(command.equals("LAB"))||(command.equals("LA"))||(command.equals("L"))))
                {
                    inputStream.mark(READ_AHEAD_BUFFER);
                    checkLabelFormat();
                }
                else if((command.equals("MODIFY"))||(command.equals("MODIF"))||(command.equals("MODI"))||(command.equals("MOD"))||(command.equals("MO"))||(command.equals("M")))
                    checkModifyFormat(currentLine);
                else if((command.equals("TITLE"))||(command.equals("TITL"))||(command.equals("TIIT"))||(command.equals("TI"))||(command.equals("T")))
                    checkTitleFormat();
                currentLine = inputStream.readLine();
            }
        }
        catch(IOException e)
        {
            this.message.add("An error occured while attempting to read the .db file.\n  " + e.getMessage());
        }
    }
    
    private String getCommand(String line) //Returns an uppercase String representation of the first tokeen in line, for keyword checks
    {
        if(line==null)
            return "";
        StringTokenizer tokenizer = new StringTokenizer(line);
        if(tokenizer.hasMoreTokens())
            return (tokenizer.nextToken()).toUpperCase();
        else            //line is whitespace
            return line;
    }
    
    private void getLabels() throws IOException
    {
        String current = inputStream.readLine();
        String command = getCommand(current);
        while((command!=null)&&!((command.equals("DATA"))||(command.equals("DAT"))||(command.equals("DA"))||(command.equals("D"))))
        {    
            current = inputStream.readLine();
            command = getCommand(current);
        }
        //Put labels into array:
        if(current!=null)
        {
            StringTokenizer tokenizer = new StringTokenizer(current);
            tokenizer.nextToken(); //"DATA"
            labelHolder = new String[tokenizer.countTokens()];
            int count = 0;
            while(count<labelHolder.length)
            {
                labelHolder[count] = tokenizer.nextToken();
                count++;
            }
        }
        inputStream.reset();
    }
    
    public String toString()      //Returns message stating the problems with the .db file.
    {
        String output = this.message.size() + " errors found.\n";
        if(this.message.size()==0)
            output = output + "There appear to be no errors in the file " + this.dbFile.getName() + ".";
        else if(this.message.size()==1)
        {
            output = "1 error found.\nThe following error appears in the file " + this.dbFile.getName() + ":\n- ";
            output = output + (String)(this.message.elementAt(0));
        }
        else
        {
            output = output + "The following errors appear in the file " + this.dbFile.getName() + ":";
            for(int i=0; i<this.message.size(); i++)
                output = output + "\n- " + (String)(this.message.elementAt(i));
        }        
        return output;
    }
    
    private void checkDataFormat(String firstLine) throws IOException // Checks formatting of DATA "block" for errors.
    {
        if((firstLine.charAt(0)=='\\')||(firstLine.length()==0))
            this.message.addElement("This .db file contains no data.");
        boolean endLoop = false; //Keeps this method from printing lots of identical messages if there is a problem.
        String currentLine = firstLine;
        int lineCount = 0;
        while(!endLoop&&(currentLine!=null))
        {
            if(hasErrors(currentLine))
            {    
                this.message.addElement("This .db file has errors in the data format around line " + lineCount + " of the DATA block.");
                endLoop = true;
            }
            lineCount++;
            currentLine = inputStream.readLine();
        }
        inputStream.reset();
    }
    
    private boolean hasErrors(String line)      //Returns true if the line contains an error; false otherwise
    {
        if(line==null)
            return false;
        else if(line.length()==0)
            return false;
        else if(line.charAt(0)=='!')      //indicates comment
            return false;
        else if(line.endsWith("\\"))
        {
            if(numberOfCommas(line,0)<3)        //Three commas for value, positive error, and negative error
                return true;
            else if(numberOfCommas(line, 0)==3)
                return false;
            else if(numberOfCommas(line, 0)>3)  //There may just be no break between columns of data
            {
                if(numberOfCommas(line, 0)%3!=0)    //There could be a partial line of data or just an out-of place comma... which would be an error
                    return true;
                //Check to see how many column headers there are. There should be commas per column header
                else
                {
                    StringTokenizer tokenizer = new StringTokenizer(line);
                    int headerCount = 0;
                    while(tokenizer.hasMoreTokens())
                        if(isHeader(tokenizer.nextToken()))
                            headerCount++;
                    if((headerCount==0)||(numberOfCommas(line,0)/headerCount==3))
                        return false;
                    else
                        return true;
                }
            }
        }
        return false;
    }
    
    private int numberOfCommas(String str, int commaCount) //Used to count commas for checkDataFormat()
    {
        int commaIndex = str.indexOf(",");
        if((commaIndex<0)||(commaIndex==str.length()-1))
            return commaCount;
        else
            return numberOfCommas(str.substring(commaIndex+1), commaCount+1);       //Uses recursion to simplify things.
    }
    
    private boolean isHeader(String token)      //called by hasErrors to check if tokens are in fact headers    
    {
        return((Character.isLetter(token.charAt(0)))&&(token.indexOf("=")>0));
    }
    
    private void checkLabelFormat() throws IOException   //Checks formatting of LABELS "block" for errors.
    {
        int labelCount = 0;             //Errors will occur if the number of labels (which label data columns) does not match the number of data columns.
        String currentLine = inputStream.readLine();
        if((currentLine==null)||(currentLine.length()==0))
            this.message.add("This .db file does not appear to contain any labels.");
        while((currentLine!=null)&&(currentLine.length()!=0))
        {
            if((currentLine.trim().length()>0)&&(currentLine.trim().charAt(0)!='!'))      //In case it is a comment
                labelCount++;
            currentLine = inputStream.readLine();
        }
        
        //Now to check how many data columns there are:
        if(labelCount!=this.labelHolder.length)
            this.message.addElement("The number of labels (" + labelCount + ") and the number of columns (" + this.labelHolder.length + ") in this .db file are different.\n  The DataTable may not load properly.");
        //Now, bring the inputStream back to where it was when this method was called.
        inputStream.reset();
        //Check to ensure that no label name occurs twice.
        if(duplicatesOccur(labelHolder))
                this.message.addElement("Two columns share the same label. This may cause ambiguities in reading and using this .db file.");
        labelsHaveBeenChecked = true;
    }
    
    private boolean duplicatesOccur(String[] label) //Used to check for duplicate labels
    {
        for(int i=0; i<label.length; i++)
            for(int j=i+1; j<label.length; j++)
                if(label[j].equals(label[i]))
                    return true;
        return false;
    }
    
    private boolean isInLabelHolder(String label)
    {
        for(int i=0; i<labelHolder.length; i++)
            if(label.equals(labelHolder[i]))
                return true;
        return false;
    }
    
    private void checkDefineFormat(String line)     //Checks formatting of DEFINE "block" for errors
    {
        StringTokenizer tokenizer = new StringTokenizer(line); //There should be four tokens: "DEFINE", the constant name, the equals sign, and the constant expression
        if(tokenizer.countTokens()!=4)
            this.message.addElement("There are errors in the syntax of a constant definition statement in this .db file.");
        tokenizer.nextToken(); //"DEFINE"
        tokenizer.nextToken(); //constant name
        if(!tokenizer.nextToken().equals("="))   //It should be the equals sign.
            this.message.addElement("Constant definition statements must have the following format: DEFINE [constant name] = [constant expression].");
    }
    
    private void checkModifyFormat(String line) throws IOException    //Checks formatting of the MODIFY "block" for errors
    {
        StringTokenizer tokenizer = new StringTokenizer(line);
        if(tokenizer.countTokens()<2)
            this.message.addElement("There are errors in the syntax of a \"modify\" statment in this .db file.\n  This statement must specify the column to be modified.");
        String nextLine = inputStream.readLine();       //This line should say how the above column is to be modified
        if((nextLine==null)||(nextLine.length()==0))
            this.message.addElement("A modify statement in this .db file does not specify the modification to be made.");
    }
    
    private void checkDeleteFormat(String firstLine) throws IOException        //Checks formatting of the DELETE "block" for errors
    {
        StringTokenizer tokenizer = new StringTokenizer(firstLine);
        if(tokenizer.countTokens()<2)
            this.message.addElement("A delete statement in this .db file does not specify the deletion to be made.");
        else
        {
            tokenizer.nextToken();      //i.e., DELETE
            String deletedColumn = tokenizer.nextToken();       //Name of column to be deleted
            if(!deletedColumn.equals("="))      //This ensure that is was not actually a column called del or d or something.
            {
                //Check if this column named in the DELETE commmand does, indeed, exist:
                if(!isInLabelHolder(deletedColumn))
                    this.message.addElement("This .db file contains a DELETE statement which attempts to delete a column which does not exist: " + deletedColumn);
            }
        }
    }
    
    private void checkGenerateFormat(String firstLine)      //Checks formatting of the GENERATE "block" for errors
    {
        StringTokenizer tokenizer = new StringTokenizer(firstLine);
        if(tokenizer.countTokens()<2)
            this.message.addElement("A generate statement in this .db file does not specify the name of the column to be generated.");
        tokenizer.nextToken(); //i.e., "GENERATE"
        String columnName = tokenizer.nextToken();
        if(isInLabelHolder(columnName))
            this.message.addElement("A generate statement in this .db file attempts to generate a column with a name that is already in use: " + columnName);
    }
    
    private void checkTitleFormat() throws IOException
    {
        String line = inputStream.readLine();
        if((line==null)||line.length()==0)
            this.message.add("This .db file does not appear to contain a title for the DataTable.");
    }
}