package musr.client.dbchecker;
import java.io.*;
import java.util.*;
import javax.swing.*;

/**
     *The methods in this class take a file and check it for errors, then attempt to correct those errors. 
     *If the program cannot guess the correct format, it puts the line into the format of a comment with 
     *two asterisks in front of it. 
     *
     *It then saves the file with corrections into a file with the word "_fixed" in the name, to distinguish 
     *the new file from the original .db file. If no corrections were made, no new file is created.
     */

public class DBFixer
{
    
    private File dbFile;                                    //File being evaluated
    private BufferedReader inputStream = null;              //To read from file
    public static final int READ_AHEAD_BUFFER = 250000;     //Buffer size for calls to inputStream.mark();
    
    private String[] labelHolder;                           //Holds column labels
    private boolean labelsHaveBeenChecked;                  //Checks whether column labels have been checked for validity
    private int labelCount;                                 //Counts labels
    private int labelsToRemove;                             //Keeps track of how many columns must be added by fixData()
    private boolean duplicateLabels;                        //Keeps track of whether there are duplicate labels
    
    private int lineCounter;                                //Keeps track of how many lines have been read in
    private int lineCounterMark;                            //For lineCounter mark, reset
    
    private Vector dataErrors;                                      //Keeps track of where errors occured in the DATA block(s)
    private Vector defineErrors;                                    //Keeps track of where errors occured in the DEFINE block(s)
    private Vector deleteErrors;                                    //Keeps track of where errors occured in the DELETE block(s)
    private Vector generateErrors;                                  //Keeps track of where errors occured in the GENERATE block(s)
    private Vector labelErrors;                                     //Keeps track of where errors occured in the LABEL block(s)
    private Vector modifyErrors;                                    //Keeps track of where errors occured in the MODIFY block(s)
    private Vector titleErrors;                                     //Keeps track of where errors occured in the TITLE block
    
    private boolean errorsInData = false;                   //Checks whether there are errors in the DATA block(s)
    private boolean errorsInDefine = false;                 //Checks whether there are errors in the DEFINE block(s)
    private boolean errorsInDelete = false;                 //Checks whether there are errors in the DELETE block(s)
    private boolean errorsInGenerate = false;               //Checks whether there are errors in the GENERATE block(s)
    private boolean errorsInLabels = false;                 //Checks whether there are errors in the LABEL block(s)
    private boolean errorsInModify = false;                 //Checks whether there are errors in the MODIFY block(s)
    private boolean errorsInTitle = false;                  //Checks whether there are errors in the TITLE block
    
    Vector buffer;                                          //Holds contents of file during fix operations
    PrintWriter outputStream;                               //Writes to file during fix operations
    
    public DBFixer()
    {
        dbFile = null;
        labelHolder = new String[0];
        labelsHaveBeenChecked = false;
        labelCount = 0;
        labelsToRemove = 0;
        duplicateLabels = false;
        lineCounter = 0;
        dataErrors = new Vector();
        defineErrors = new Vector();
        deleteErrors = new Vector();
        generateErrors = new Vector();
        labelErrors = new Vector();
        modifyErrors = new Vector();
        titleErrors = new Vector();
        buffer = new Vector();
    }
    
    public DBFixer(File newDB)
    {
        dbFile = newDB;
        labelHolder = new String[0];
        labelsHaveBeenChecked = false;
        labelCount = 0;
        labelsToRemove = 0;
        duplicateLabels = false;
        lineCounter = 0;
        dataErrors = new Vector();
        defineErrors = new Vector();
        deleteErrors = new Vector();
        generateErrors = new Vector();
        labelErrors = new Vector();
        modifyErrors = new Vector();
        titleErrors = new Vector();
        buffer = new Vector();
    }
    
    public void fix()
    {
        //System.out.println("Fix beginning.");
        if(checkFile())
        {
           //System.out.println("Content check beginning.");
           checkContents();
        }
        else
            JOptionPane.showMessageDialog(null, dbFile.getName() + "\nThis program was unable to fix the file, as it was not a valid, readable .db file.");
        System.exit(0);
    }
    
    public void setFile(File newDB)
    {
        dbFile = newDB;
    }
    
    public void reset()
    {
        dbFile = null;
        labelHolder = new String[0];
        labelsHaveBeenChecked = false;
        labelCount = 0;
        labelsToRemove = 0;
        duplicateLabels = false;
        lineCounter = 0;
        dataErrors.clear();
        defineErrors.clear();
        deleteErrors.clear();
        generateErrors.clear();
        labelErrors.clear();
        modifyErrors.clear();
        titleErrors.clear();
        buffer.clear();
        errorsInData = false;
        errorsInDefine = false;
        errorsInDelete = false;
        errorsInGenerate = false;
        errorsInLabels = false;
        errorsInModify = false;
        errorsInTitle = false;
    }
    
    public boolean checkFile()      //Returns true if file has been repaired to be readable .db file. Returns false otherwise.
    {
        if((this.dbFile == null)||(this.dbFile.length()==0))
            return false;
        if(!this.dbFile.exists())
            return false;
        if(!this.dbFile.canRead())
            return false;
        if(!hasDBExtension())
        {    
            if(fixExtension())    
                return true;
            else
                return false;
        }
        return true;
    }
    
    private boolean hasDBExtension()        //Checks if file's extension is .db.
    {
        String filename = this.dbFile.getName();
        filename = filename.toLowerCase();
        return(filename.indexOf(".db")>0);
    }
    
    private boolean fixExtension()          //Changes file's extension to .db.
    {
        String filepath = this.dbFile.getPath();
        int extensionIndex = filepath.indexOf(".");
        if(extensionIndex>0)
        {
            filepath = filepath.substring(0, extensionIndex);
            return this.dbFile.renameTo(new File(filepath + ".db"));
        }
        else
            return false;
    }
    
    public void checkContents()         //Checks the contents of the filxe
    {
        try
        {
            inputStream = new BufferedReader(new FileReader(dbFile));
            inputStream.mark(READ_AHEAD_BUFFER);
            markLineCounter();
            getLabels();
            String currentLine = inputStream.readLine();
            //System.out.println("Stream opened.");
            if(currentLine==null)
                JOptionPane.showMessageDialog(null, "The file " + dbFile.getName() + " is empty.");
            String command;
            while(currentLine!=null)
            {   //Enlarge all Vectors to give them room for potential new Boolean(true) values:
                if(lineCounter+1>dataErrors.size())
                    this.dataErrors.setSize(lineCounter+1);
                this.defineErrors.setSize(lineCounter+1);
                this.deleteErrors.setSize(lineCounter+1);
                this.generateErrors.setSize(lineCounter+1);
                if(lineCounter+1>labelErrors.size())
                    this.labelErrors.setSize(lineCounter+1);
                if(lineCounter+1>modifyErrors.size())
                    this.modifyErrors.setSize(lineCounter+1);
                this.titleErrors.setSize(lineCounter+1);
                
                command = getCommand(currentLine);
                if((command.equals("DATA"))||(command.equals("DAT"))||(command.equals("DA"))||(command.equals("D")))
                {
                    inputStream.mark(READ_AHEAD_BUFFER);
                    markLineCounter();
                    checkDataFormat(currentLine);
                }
                else if((command.equals("DEFINE"))||(command.equals("DEFIN"))||(command.equals("DEFI"))||(command.equals("DEF")))
                    checkDefineFormat(currentLine);
                else if((command.equals("DELETE"))||(command.equals("DELET"))||(command.equals("DELE"))||(command.equals("DEL")))
                    checkDeleteFormat(currentLine);
                else if((command.equals("GENERATE"))||(command.equals("GENERAT"))||(command.equals("GENERA"))||(command.equals("GENER"))||(command.equals("GENE"))||(command.equals("GEN"))||(command.equals("GE"))||(command.equals("G")))
                    checkGenerateFormat(currentLine);
                else if((!labelsHaveBeenChecked)&&((command.equals("LABELS"))||(command.equals("LABEL"))||(command.equals("LABE"))||(command.equals("LAB"))||(command.equals("LA"))||(command.equals("L"))))
                {
                    inputStream.mark(READ_AHEAD_BUFFER);
                    markLineCounter();
                    checkLabelFormat();
                }
                else if((command.equals("MODIFY"))||(command.equals("MODIF"))||(command.equals("MODI"))||(command.equals("MOD"))||(command.equals("MO"))||(command.equals("M")))
                    checkModifyFormat(currentLine);
                else if((command.equals("TITLE"))||(command.equals("TITL"))||(command.equals("TIIT"))||(command.equals("TI"))||(command.equals("T")))
                    checkTitleFormat();
                
                currentLine = inputStream.readLine();
                this.lineCounter++;
            }
            inputStream.close();
            //System.out.println("Input stream closed.");
        }
        catch(IOException ioe)
        {
            JOptionPane.showMessageDialog(null, "An error occured while trying to read from file " + dbFile.getName() + ": " + ioe.getMessage());
        }
        //System.out.println("Content fix beginning.");
        fixContents();
    }
    
    private String getCommand(String line) //Returns an uppercase String representation of the first tokeen in line, for keyword checks
    {
        if(line==null)
            return "";
        StringTokenizer tokenizer = new StringTokenizer(line);
        if(tokenizer.hasMoreTokens())
            return (tokenizer.nextToken()).toUpperCase();
        else            //line is whitespace
            return line;
    }
    
    private void getLabels() throws IOException         //Puts column labels into array
    {
        String current = inputStream.readLine();
        String command = getCommand(current);
        while((command!=null)&&!((command.equals("DATA"))||(command.equals("DAT"))||(command.equals("DA"))||(command.equals("D"))))
        {    
            current = inputStream.readLine();
            lineCounter++;
            command = getCommand(current);
        }
        //Put labels into array:
        if(current!=null)
        {
            StringTokenizer tokenizer = new StringTokenizer(current);
            tokenizer.nextToken(); //"DATA"
            labelHolder = new String[tokenizer.countTokens()];
            int count = 0;
            while(count<labelHolder.length)
            {
                labelHolder[count] = tokenizer.nextToken();
                count++;
            }
        }
        inputStream.reset();
        resetLineCounter();
    }
    
    private void checkDataFormat(String line) throws IOException      //Checks the format of the data.
    {
        if((line.charAt(0)=='\\')||(line.length()==0))
            JOptionPane.showMessageDialog(null, "The .db file " + this.dbFile.getName() + " contains no data.");
        String currentLine = line;
        while(currentLine!=null)
        {
            if(hasErrors(currentLine))    
                this.dataErrors.setElementAt(new Boolean(true), this.lineCounter);
            currentLine = inputStream.readLine();
            lineCounter++;
            this.dataErrors.setSize(lineCounter+1);
        }
        inputStream.reset();
        resetLineCounter();
    }
    
    private boolean hasErrors(String line)      //Returns true if the line contains an error; false otherwise
    {
        if(line==null)
            return false;
        else if(line.length()==0)
            return false;
        else if(line.charAt(0)=='!')      //indicates comment
            return false;
        else if(line.endsWith("\\"))
        {
            if(numberOfCommas(line,0)<3)        //Three commas for value, positive error, and negative error
                return true;
            else if(numberOfCommas(line, 0)==3)
                return false;
            else if(numberOfCommas(line, 0)>3)  //There may just be no break between columns of data
            {
                if(numberOfCommas(line, 0)%3!=0)    //There could be a partial line of data or just an out-of place comma... which would be an error
                    return true;
                //Check to see how many column headers there are. There should be three commas per column header
                else
                {
                    StringTokenizer tokenizer = new StringTokenizer(line);
                    int headerCount = 0;
                    while(tokenizer.hasMoreTokens())
                        if(isHeader(tokenizer.nextToken()))
                            headerCount++;
                    if((headerCount==0)||(numberOfCommas(line,0)/headerCount==3))
                        return false;
                    else
                        return true;
                }
            }
        }
        return false;
    }
    
    private int numberOfCommas(String str, int commaCount) //Used to count commas for checkDataFormat()
    {
        int commaIndex = str.indexOf(",");
        if(commaIndex<0)
            return commaCount;
        if(commaIndex==str.length()-1)
            return commaCount+1;
        else
            return numberOfCommas(str.substring(commaIndex+1), commaCount+1);       //Uses recursion to simplify things.
    }
    
    private boolean isHeader(String token)      //called by hasErrors to check if tokens are in fact headers    
    {
        return((Character.isLetter(token.charAt(0)))&&(token.indexOf("=")>0));
    }
    
    private void checkDefineFormat(String line)     //Checks formatting of DEFINE "block" for errors
    {
        StringTokenizer tokenizer = new StringTokenizer(line); //There should be four tokens: "DEFINE", the constant name, the equals sign, and the constant expression
        if(tokenizer.countTokens()!=4)
        {    
            errorsInDefine = true;
            defineErrors.setElementAt(new Boolean(true), this.lineCounter);
        }
        else
        {
            tokenizer.nextToken(); //"DEFINE"
            tokenizer.nextToken(); //constant name
            if(!tokenizer.nextToken().equals("="))   //It should be the equals sign.
            {    
                errorsInDefine = true;
                defineErrors.setElementAt(new Boolean(true), this.lineCounter);
            }
        }
    }
    
    private void checkDeleteFormat(String line) throws IOException        //Checks formatting of the DELETE "block" for errors
    {
        StringTokenizer tokenizer = new StringTokenizer(line);
        if(tokenizer.countTokens()<2)
        {
            this.errorsInDelete = true;
            deleteErrors.setElementAt(new Boolean(true), lineCounter);
        }
        else
        {
            tokenizer.nextToken();      //i.e., DELETE
            String deletedColumn = tokenizer.nextToken();       //Name of column to be deleted
            if(!deletedColumn.equals("="))      //This ensure that is was not actually a column called del or d or something.
            {
                //Check if this column named in the DELETE commmand does, indeed, exist:
                if(!isInLabelHolder(deletedColumn))
                {
                    this.errorsInDelete = true;
                    deleteErrors.setElementAt(new Boolean(true), this.lineCounter);
                }
            }
        }
    }
    
    private boolean isInLabelHolder(String label)           //Checks if a String is in labelHolder[]
    {
        for(int i=0; i<labelHolder.length; i++)
            if(label.equals(labelHolder[i]))
                return true;
        return false;
    }
    
    private void checkGenerateFormat(String firstLine)      //Checks formatting of the GENERATE "block" for errors
    {
        StringTokenizer tokenizer = new StringTokenizer(firstLine);
        if(tokenizer.countTokens()<2)
        {
            this.errorsInGenerate = true;
            generateErrors.setElementAt(new Boolean(true), lineCounter);
        }     
        else if(tokenizer.countTokens()==2)
        {
            tokenizer.nextToken(); //i.e., "GENERATE"
            String columnName = tokenizer.nextToken();
            if(isInLabelHolder(columnName))
            {
                this.errorsInGenerate = true;
                generateErrors.setElementAt(new Boolean(true), lineCounter);
            }
        }
    }
    
    private void checkLabelFormat() throws IOException      //Checks formatting of the LABEL block for errors
    {
        String currentLine = inputStream.readLine();
        lineCounter++;
        if((currentLine==null)||(currentLine.length()==0))
        {
            errorsInLabels = true;
            labelErrors.setElementAt(new Boolean(true), lineCounter-1);
        }
        while((currentLine!=null)&&(currentLine.length()!=0))
        {
            if((currentLine.trim().length()>0)&&(currentLine.trim().charAt(0)!='!'))      //In case it is a comment
                labelCount++;
            currentLine = inputStream.readLine();
            lineCounter++;
            this.labelErrors.setSize(this.lineCounter+1);
        }
        
        //Now to check how many data columns there are:
        
        getLabels();
        if(labelCount!=this.labelHolder.length)
        {
            errorsInLabels = true;
            labelErrors.setElementAt(new Boolean(true), lineCounter);
        }
        //Now, bring the inputStream back to where it was when this method was called.
        inputStream.reset();
        resetLineCounter();
        //Check to ensure that no label name occurs twice.
        if(duplicatesOccur(labelHolder))
        {
            errorsInLabels = true;
            labelErrors.setElementAt(new Boolean(true), lineCounter);
            this.duplicateLabels=true;
        }
        labelsHaveBeenChecked = true;
    }
    
    private boolean duplicatesOccur(String[] label) //Used to check for duplicate labels
    {
        for(int i=0; i<label.length; i++)
            for(int j=i+1; j<label.length; j++)
                if((label[i]!=null)&&(label[j]!=null)&&(label[j].equals(label[i])))
                    return true;
        return false;
    }
    
    private void checkModifyFormat(String line) throws IOException      //Checks format of MODIFY block for errors
    {
        StringTokenizer tokenizer = new StringTokenizer(line);
        if(tokenizer.countTokens()<2)
        {
            this.errorsInModify = true;
            modifyErrors.setElementAt(new Boolean(true), lineCounter);
        }
        else
        {
            String nextLine = inputStream.readLine();       //This line should say how the above column is to be modified
            lineCounter++;
            while((nextLine!=null)&&(nextLine.length()>0)&&(nextLine.charAt(0)=='!'))
            {
                nextLine = inputStream.readLine();
                lineCounter++;
            }
            if((nextLine!=null)&&(nextLine.length()>0))
            {
                nextLine = inputStream.readLine();
                lineCounter++;
            }
            modifyErrors.setSize(lineCounter+1);
            if((nextLine==null)||(nextLine.length()==0)||(nextLine.trim().length()==0))
            {
                this.errorsInModify = true;
                modifyErrors.setElementAt(new Boolean(true), lineCounter);
            }
        }
    }
    
    private void checkTitleFormat() throws IOException                  //Checks format of TITLE block for errors
    {
       String line = inputStream.readLine();
       this.lineCounter++;
       while((line.length()>0)&&(line.charAt(0)=='!'))
       {
           line = inputStream.readLine();
           lineCounter++;
       }
       titleErrors.setSize(lineCounter+1);
       if((line==null)||line.length()==0)
       {
          this.errorsInTitle = true;
          titleErrors.setElementAt(new Boolean(true), lineCounter);
       }
    }
    
    private void markLineCounter()                                      //Marks current position of lineCounter
    {
        this.lineCounterMark = this.lineCounter;
    }
    
    private void resetLineCounter()                                     //Returns to previous position of lineCounter
    {
        this.lineCounter = this.lineCounterMark;
    }
    
    private void fixContents()                    //Calls the various fix methods (below) and rewrites the file to incorporate the changes made
    {
        if(this.errorsInData||this.errorsInDefine||this.errorsInDelete||this.errorsInGenerate||this.errorsInLabels||this.errorsInModify||this.errorsInTitle)
        {
            this.buffer = new Vector();
            String currentLine;
            try
            {
                this.inputStream = new BufferedReader(new FileReader(this.dbFile));
                currentLine = inputStream.readLine();
                while(currentLine!=null)        //Fill up this.buffer
                {
                    this.buffer.add(currentLine);
                    currentLine = inputStream.readLine();
                }
                //System.out.println("Buffer readin completed.");
                renameDoubleColumns();
                //System.out.println("Double columns renamed.");
                inputStream.close();
                //System.out.println("Buffer readin stream closed.");
                
                inputStream = new BufferedReader(new FileReader(this.dbFile));
                inputStream.mark(READ_AHEAD_BUFFER);
                getLabels();
                //System.out.println("labelHolder updated");
                removeExtraLabels();
                //System.out.println("Extra labels removed.");
                inputStream.close();
                
                buffer.trimToSize();
                //Find changes, write them to file:
                for(int i=0; i<buffer.size(); i++)
                    buffer.setElementAt(fixLine(i), i);
                //Write the buffer to file:
                String filepath = this.dbFile.getPath();
                filepath = filepath.toLowerCase();
                    filepath = filepath.substring(0, filepath.indexOf(".db")) + "_fixed" + filepath.substring(filepath.indexOf(".db"));     //Save it to a slightly different name (eg DataTable_fixed.db), as this program might mangle it
                File outputFile = new File(filepath);
                while(outputFile.exists())      //Just to be sure it doesn't already exist, keep adding 0's (eg DataTable_fixed0.db)
                {
                    filepath = filepath.substring(0, filepath.indexOf(".db")) + "0" + filepath.substring(filepath.indexOf(".db"));
                    outputFile = new File(filepath);
                }
                this.outputStream = new PrintWriter(new FileOutputStream(outputFile));
                for(int i=0; i<buffer.size(); i++)
                    outputStream.println((String)buffer.elementAt(i));
                outputStream.close();
                JOptionPane.showMessageDialog(null, dbFile.getName() + "\nA new file with fixes was created at " + filepath + ".");
            }
            catch(IOException ioe)
            {
                JOptionPane.showMessageDialog(null, "An error occured during file i/o: " + ioe.getMessage());
            }
        }
        else
            JOptionPane.showMessageDialog(null, "No errors were found in " + dbFile.getName() + ".");
    }
    
    private void removeExtraLabels()            //Removes "extra" labels - duplicates, or labels in excess of the number of columns
    {        
        this.labelsToRemove = labelCount - labelHolder.length - 1;
        
        int index = 0;
        String command = getCommand((String)this.buffer.elementAt(index));
        while((index<this.buffer.size())&&(!command.equals("LABELS"))&&(!command.equals("LABEL"))&&(!command.equals("LABE"))&&(!command.equals("LAB"))&&(!command.equals("LAB"))&&(!command.equals("LA"))&&(!command.equals("L")))
        {
            command = getCommand((String)this.buffer.elementAt(index));
            index++;
        }
        int labelsStart = index;
        
        if(this.duplicateLabels)
        {
            String[] duplicateFinder = new String[labelHolder.length];
            for(int i=0; i<labelHolder.length; i++)
            {
                duplicateFinder[i] = labelHolder[i];
                if(duplicatesOccur(duplicateFinder))
                {
                    this.buffer.remove(duplicateFinder[i]);
                    labelsToRemove--;
                    duplicateFinder[i] = "supercallifragilisticexpialidoucious"; //So that array entries like "" are not mistaken for duplicates, set this to something which is not likely to have a duplicate
                }
            }
        }
        
    //Just arbirtarily take some labels out if necessary:
        for(int i=1; i<labelsToRemove; i++)
            this.buffer.remove(labelsStart+i);
    }
    
    private void renameDoubleColumns()
    {
        int index = 0;
        String currentLine = (String)this.buffer.elementAt(index);
        while((index<this.buffer.size())&&(!(getCommand(currentLine).equals("DATA"))||(getCommand(currentLine).equals("DAT"))||(getCommand(currentLine).equals("DA"))||(getCommand(currentLine).equals("D"))))
        {
            index++;
            currentLine = (String)this.buffer.elementAt(index);
        }
        //Now that we have found the line with the data in it, find the column headers and make sure that none of them are the same:
        StringTokenizer tokenizer = new StringTokenizer(currentLine);
        tokenizer.nextToken();  //Discard DATA
        String[] header = new String[tokenizer.countTokens()];
        for(int i=0; i<header.length; i++)
            header[i] = tokenizer.nextToken();
        int j=0;
        String[] duplicateFinder = new String[header.length];       //Copy header into new array element by element, looking for duplicates
        while((j<header.length)&&(duplicatesOccur(header)))
        {
            duplicateFinder[j] = header[j];
            if(duplicatesOccur(duplicateFinder))
                duplicateFinder[j] = duplicateFinder[j] + "0";
            j++;
        }
        //Finish filling duplicateFinder, if necessary:
        for(int i=0; i<duplicateFinder.length; i++)
            if(duplicateFinder[i]==null)
                duplicateFinder[i] = header[i];
        String result = "DATA ";
        for(int i=0; i<duplicateFinder.length; i++)
            result = result + duplicateFinder[i] + " ";
        this.buffer.setElementAt(result, index);
    }
    
    private String fixLine(int index)               //Corrects errors in the line at index
    {
        String line = (String)buffer.elementAt(index);
        if((this.dataErrors.elementAt(index)!=null)&&(((Boolean)this.dataErrors.elementAt(index)).booleanValue()))
        //{    System.out.println("data " + index); 
            line = fixData(line);//}
        else if((this.defineErrors.elementAt(index)!=null)&&(((Boolean)this.defineErrors.elementAt(index)).booleanValue()))
        //{   System.out.println("define " + index);
            line = fixDefine(line);//}
        else if((this.deleteErrors.elementAt(index)!=null)&&(((Boolean)this.deleteErrors.elementAt(index)).booleanValue()))
        //{System.out.println("delete " + index);
            line = fixDelete(line);//}
        else if((this.generateErrors.elementAt(index)!=null)&&(((Boolean)this.generateErrors.elementAt(index)).booleanValue()))
        //{   System.out.println("generate " + index);
            line = fixGenerate(line);//}
        else if((this.labelErrors.elementAt(index)!=null)&&(((Boolean)this.labelErrors.elementAt(index)).booleanValue()))
        //{   System.out.println("labels " + index);    
            line = fixLabels(line, index);//}
        else if((this.modifyErrors.elementAt(index)!=null)&&(((Boolean)this.modifyErrors.elementAt(index)).booleanValue()))
        //{   System.out.println("modify " + index);
            line = fixModify(line, index);//}
        else if((this.titleErrors.elementAt(index)!=null)&&(((Boolean)this.titleErrors.elementAt(index)).booleanValue()))
        //{   System.out.println("title " + index);
            line = fixTitle();//}
        return line;
    }
    
    private String fixData(String line)
    {
        int commaCount = numberOfCommas(line, 0);
        int slashIndex = line.indexOf("\\");        //Line must end in slash
        if(commaCount==1)       //No commas for positive, negative error, perhaps
            return line.substring(0, slashIndex) + ", , " + line.substring(slashIndex);
        else if(commaCount==2)  //Errors are the same, and no extra comma was put in, perhaps
            return line.substring(0, slashIndex) + ", " + line.substring(slashIndex);
        else if(commaCount%3==0)        //Line lacks at least one label
        {
            StringTokenizer tokenizer = new StringTokenizer(line);
            String currentToken;
            int headerCount = 0;
            if(tokenizer.countTokens()>0)
            { 
                currentToken = tokenizer.nextToken();
                while(tokenizer.hasMoreTokens())
                {    
                    currentToken = tokenizer.nextToken();
                    headerCount++;
                }
                if(headerCount*3+3 == commaCount)        //One more label needed
                    return labelHolder[0] + line;
                else        //Comment out error within line of data
                    return "!** " + line;
            }
            else    //empty line
                return line;
        }
        else    //There is an abmiguous error within a line of data. Comment it out.
            return "!** " + line;
    }
    
    private String fixDefine(String line)
    {
        StringTokenizer tokenizer = new StringTokenizer(line);
        String constant;
        String thirdToken;
        
        if(tokenizer.countTokens()<2)         //Not even a constant name is specified; comment this out
            return"!** " + line;
        else if(tokenizer.countTokens()<3)    //Check if token after DEFINE could be a constant name
        {
            tokenizer = new StringTokenizer(line);
            tokenizer.nextToken();          //DEFINE
            constant = tokenizer.nextToken();
            if(Character.isLetter(constant.charAt(0)))  //It could be a constant name
                return line + " = -999";                 //Arbitrarily
            else                            //comment it out
                return "!** " + line;
        }
        else if(tokenizer.countTokens()<4)      //Check what's missing - equals sign, or value?
        {
            tokenizer = new StringTokenizer(line);
            tokenizer.nextToken();
            constant = tokenizer.nextToken();
            thirdToken = tokenizer.nextToken().trim();
            if(thirdToken.equals("="))          //All that's missing is the value
                return line + " =-999";         //Arbitrarily
            else if((Character.isDigit(thirdToken.charAt(0)))||(thirdToken.charAt(0)=='-')) //This is the constant, so supply the equals sign
                return "DEFINE " + constant + " = " + thirdToken;
            else        //Something is wrong; comment it out
                return"!** " + line;
        }
        else        //Something is wrong; comment it out
            return "!** " + line;
    }
    
    private String fixDelete(String line)
    {
        //Either there is no column specified for the deletion, or the specified column does not exist. Either way, just comment it out.
        return "!** " + line;
    }
    
    private String fixGenerate(String line)
    {
        StringTokenizer tokenizer = new StringTokenizer(line);
        if(tokenizer.countTokens()<2)       //No new column specified, so comment this out.
            return "!** " + line;
        else        //Trying to generate a column with an already-taken name, so change the name slightly.
        {
            String result;
            result = tokenizer.nextToken();     //GENERATE
            result = result + " " + tokenizer.nextToken() + "00";   //change the name
            while(tokenizer.hasMoreTokens())
                result = result + " " + tokenizer.nextToken();
            return result;
        }
    }
    
    private String fixLabels(String line, int index)
    {
        if(((String)this.buffer.elementAt(index+1))==null)      //The kind of error where there are no labels under the heading LABELS
            return "!** " + line;
        else if(this.labelHolder.length>this.labelCount)        //Add more labels
        {
            line = "LABELS\nDummyLabel 0";
            for(int i=1; i<labelHolder.length-labelCount; i++)
                line = line + "\nDummyLabel " + i;
            return line;
        }
        else if(this.labelHolder.length<this.labelCount)        //Take away labels
        {
            this.labelsToRemove = labelCount - labelHolder.length - 1;
            return "";
        }
        else        //There are duplicate labels; this is already dealt with
        {
            return "";
        }
    }
    
    private String fixModify(String line, int index)
    {
        if(getCommand(line).indexOf("M")==0)       //The line starts with MODIFY and thus is the kind of error where no parameters were specified
            return "!** " + line;                   //Comment out this line
        else   //The kind of error where there are no parameters for modification
        {
            String prevLine = (String)this.buffer.elementAt(index-1);
            StringTokenizer tokenizer = new StringTokenizer(prevLine);
            tokenizer.nextToken();      //Throw out MODIFY
            String result = "1*";
            while(tokenizer.hasMoreTokens())    //Put them into the MODIFY statement
                result = result + tokenizer.nextToken() + " ";
            return result;
        }
    }
    
    private String fixTitle()
    {
        return "[untitled]";
    }
}