package musr.client.gui;
import java.util.Vector;
import musr.client.datacolumn.*;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.text.DecimalFormat;


public class ShowStatistics extends JFrame implements java.io.Serializable, ActionListener
{
    private Frame parent;                       //Calling object
    public static final int HEIGHT = 325;       //Height of main GUI
    public static final int WIDTH = 510;        //Width of main GUI
    private Statistics stats;                   //The Statistics object doing the calculations
    private String columnTitle;                 //Title of selected column in calling object
    private DecimalFormat largeNumberFormatter; //Formats numbers x>=1e9
    private DecimalFormat mediumNumberFormatter;//Formats numbers 1e-5<=x<1e9
    private DecimalFormat smallNumberFormatter; //Formats numbers x<1e-5
    
    /**
     Constructs a new ShowStatistics object. 
     @param parentFrame The GUI from which this constructor is called. The Vector holding the values of the selected column is used to create a new Statistics object.
     */
    public ShowStatistics(Frame parentFrame)
    {
        this.setSize(WIDTH, HEIGHT);
        
        Container contentPane = this.getContentPane();
        contentPane.setLayout(new BorderLayout());
        //Copy the data from the selected column into a Statistics object:
        int dataTableKey = ((GUI)parentFrame).getDataTableIndex();
        Dictionary temp = ((GUI)parentFrame).dictionary;
        DataTable selectedDataTable;
        if((temp!=null)&&(dataTableKey<temp.size()))
            selectedDataTable = (DataTable)temp.getElement(dataTableKey);
        else
            selectedDataTable = new DataTable("/tmp/default.gui");
        Vector values = new Vector();  //Keeps compiler happy to declare outside of if statement
        int columnIndex=0;      //Keeps compiler happy to declare outside of if statement
        if(selectedDataTable.columnCount()!=0)
        {
            columnIndex = ((GUI)parentFrame).columnList.getSelectedIndex();
            for(int i=0; i<selectedDataTable.getColumn(columnIndex).getSize(); i++)
                values.add(selectedDataTable.getColumn(columnIndex).getValue(i));
        }
        else
            ((GUI)parentFrame).statusTextField.setText("The DataTable contains no columns.");
        stats = new Statistics(values);
        
        columnTitle = selectedDataTable.getColumn(columnIndex).getLabel(); 
        this.setTitle("Column: " + columnTitle);
        //Set up the formatters for various-sized numbers (will make them print nicely)
        largeNumberFormatter = new DecimalFormat("0.0###E0");
        mediumNumberFormatter = new DecimalFormat("######0.0###");
        smallNumberFormatter = new DecimalFormat("0.0##E0");
        
        stats.calculate();      //Calculates the statistics
        //Now to put things in the GUI:
        JPanel dataPane = new JPanel();
        dataPane.setLayout(new GridLayout(13, 2));
        
        dataPane.add(new JLabel("Number of values:"));
        dataPane.add(new JLabel("   " + stats.getSize()));
        dataPane.add(new JLabel("Maximum value:"));
        dataPane.add(new JLabel("   " + format(stats.getMax())));
        dataPane.add(new JLabel("Minimum value:"));
        dataPane.add(new JLabel("   " + format(stats.getMin())));
        dataPane.add(new JLabel("Median value:"));
        dataPane.add(new JLabel("   " + format(stats.getMedian()) + ""));      
        dataPane.add(new JLabel("Mean:"));
        dataPane.add(new JLabel(""));   //Blank space
        dataPane.add(new JLabel("  -Including outliers:"));
        dataPane.add(new JLabel("   " + format(stats.getMean())));
        dataPane.add(new JLabel("  -Outliers removed:"));
        dataPane.add(new JLabel("   " + format(stats.getMeanWithoutOutliers())));
        dataPane.add(new JLabel("Standard deviation:"));
        dataPane.add(new JLabel(""));   //blank space
        dataPane.add(new JLabel("  -Including outliers:"));
        dataPane.add(new JLabel("   " + format(stats.getStdDev())));
        dataPane.add(new JLabel("  -Outliers removed:"));
        dataPane.add(new JLabel("   " + format(stats.getStdDevWithoutOutliers())));
        dataPane.add(new JLabel("Skewness:"));
        dataPane.add(new JLabel(""));   //blank space
        dataPane.add(new JLabel("  -Standard skewness:"));
        dataPane.add(new JLabel("   " + format(stats.getSkewness())));
        dataPane.add(new JLabel("  -Bowley skewness:"));
        dataPane.add(new JLabel("   " + format(stats.getBowleySkewness())));
        
        dataPane.setBorder(BorderFactory.createTitledBorder("Statistics"));
        contentPane.add(dataPane, BorderLayout.WEST);
        
        JPanel buttonsPane = new JPanel();
        buttonsPane.setLayout(new GridLayout(6,1));
        
        JButton pValue = new JButton("P-value");
        pValue.setToolTipText("The probability that a randomly chosen value in the set is less than the selected value.");
        pValue.addActionListener(this);
        buttonsPane.add(pValue);
        
        JButton percentile = new JButton("Percentile");
        percentile.setToolTipText("The proportion of values in this set with values less than the selected value.");
        percentile.addActionListener(this);
        buttonsPane.add(percentile);
        
        JButton relFreq = new JButton("Relative Frequency");
        relFreq.setToolTipText("The proportion of values in the set lying between two selected values (inclusive).");
        relFreq.addActionListener(this);
        buttonsPane.add(relFreq);
        
        JButton outliers = new JButton("Outliers");
        relFreq.setToolTipText("The value statistically deemed to be outliers from the rest of the set.");
        outliers.addActionListener(this);
        buttonsPane.add(outliers);
        
        buttonsPane.add(new JLabel(""));   //Blank space
        
        JButton close = new JButton("Close");
        close.setToolTipText("Close the data statistics window.");
        close.addActionListener(this);
        buttonsPane.add(close);
        
        contentPane.add(buttonsPane, BorderLayout.EAST);
    }
    //Makes numbers print nicely and legibly, and prevents any blanks.
    private String format(double number)
    {
        double temp = Math.abs(number); //So that formatting works regardless of sign
        if(temp==0)
            return "0";
        else if(Double.isNaN(temp))
            return "--error--";
        else if(Math.log(temp)/Math.log(10)<-4)
            return smallNumberFormatter.format(number);
        else if(Math.log(temp)/Math.log(10)<7)
            return mediumNumberFormatter.format(number);
        else
            return largeNumberFormatter.format(number);
    }
    
    /**
     Overrides method actionPerformed in interface ActionListener. Handles events generated when buttons are clicked.
     */
    public void actionPerformed(ActionEvent e)
    {
        String command = e.getActionCommand();
        if(command.equals("Close"))
            this.dispose();
        else if(command.equals("P-value"))
            showPValueDialog();
        else if(command.equals("Percentile"))
            showPercentileDialog();
        else if(command.equals("Relative Frequency"))
            showRelativeFrequency();
        else if(command.equals("Outliers"))
            showOutliers();
        else
            System.out.println("Error in user interface.");
    }
    
    /**
     Shows dialog to calculate p-values.
     @see Statistics#pValue(double)
     */
    public void showPValueDialog()
    {
        JFrame0 dialog = new JFrame0();
        dialog.setSize(800, 125);
        dialog.setTitle("P-value");
        Container contentPane = dialog.getContentPane();
        contentPane.setLayout(new BorderLayout());
        contentPane.add(new JTextArea("The P-value is the probability that a randomly selected value from the series would be greater than or equal to the given value." +
                "\nEnter a value, then click Display to find its P-value:"), BorderLayout.NORTH);
        
        dialog.textField = new JTextField("", 20);
        dialog.textField.addActionListener(dialog);
        contentPane.add(dialog.textField, BorderLayout.WEST);
        
        dialog.textArea = new JTextArea(2, 35);
        dialog.textArea.setText("No value entered yet.");
        contentPane.add(dialog.textArea, BorderLayout.EAST);
        
        JPanel buttonPanel = new JPanel();
        buttonPanel.setLayout(new FlowLayout());
        
        JButton display = new JButton("Display");
        display.addActionListener(dialog);
        buttonPanel.add(display);
        
        JButton close = new JButton("Close");
        close.addActionListener(dialog);
        buttonPanel.add(close);
        
        contentPane.add(buttonPanel, BorderLayout.SOUTH);
        
        dialog.setVisible(true);
    }
    
    /**
     Shows dialog to calculate relative frequencies.
     @see Statistics#relativeFrequency(double, double)
     */
    public void showRelativeFrequency()
    {
        JFrame1 dialog = new JFrame1();
        dialog.setSize(720, 150);
        dialog.setTitle("Relative Frequency");
        
        Container contentPane = dialog.getContentPane();
        contentPane.setLayout(new BorderLayout());
        contentPane.add(new JLabel("Enter upper and lower limits, and click Display to see the proportion of values which lie at or between these limits."), 
                BorderLayout.NORTH);
    
        JPanel dialogPanel = new JPanel();
        dialogPanel.setLayout(new BorderLayout());
        
        JPanel inputPanel = new JPanel();
        inputPanel.setLayout(new GridLayout(2,2));
        inputPanel.add(new JLabel(""));      //Blank space     
        inputPanel.add(new JLabel("Lower limit:"));
        inputPanel.add(new JLabel("Upper limit:"));
        inputPanel.add(new JLabel(""));      //Blank space   
        inputPanel.add(new JLabel(""));     //Blank space       
        dialog.minField = new JTextField();
        inputPanel.add(dialog.minField);
        dialog.maxField = new JTextField();
        inputPanel.add(dialog.maxField);
        inputPanel.add(new JLabel(""));     //Blank space
        
        dialogPanel.add(inputPanel, BorderLayout.NORTH);
        dialog.output = new JLabel();
        dialogPanel.add(dialog.output, BorderLayout.SOUTH);
        
        contentPane.add(dialogPanel, BorderLayout.CENTER);
        
        JPanel buttonPanel = new JPanel();
        buttonPanel.setLayout(new FlowLayout());
        
        JButton display = new JButton("Display");
        display.setToolTipText("See the proportion of values between the given limits.");
        display.addActionListener(dialog);
        buttonPanel.add(display);
        
        JButton close = new JButton("Close");
        close.setToolTipText("Close this statistics dialog.");
        close.addActionListener(dialog);
        buttonPanel.add(close);
        
        contentPane.add(buttonPanel, BorderLayout.SOUTH);
        
        dialog.setVisible(true);
    }
    
    /**
    Displays outlying values.
     @see Statistics#outliers()
     */
    public void showOutliers()
    {
        JFrame2 dialog = new JFrame2();
        dialog.setSize(620,125);
        JPanel outputFrame = new JPanel();
        Container contentPane = dialog.getContentPane();
        contentPane.setLayout(new BorderLayout());
        
        String output = "The following ";
        double[] outlierHolder = stats.outliers();
        if(outlierHolder.length==0)
            output = "No outlying values were found in the selected column.";
        else if(outlierHolder.length==1)
            output = output + "value has been identified as the outlying value of the selected data column:";
        else
            output = output + "values have been identified as the outlying values of the selected data column: ";
        for(int i=0; i<outlierHolder.length; i++)
            output = output + "\n" + format(outlierHolder[i]);
        outputFrame.add(new JTextArea(output));
        
        JScrollPane scroller = new JScrollPane(outputFrame);
        contentPane.add(scroller, BorderLayout.CENTER);
        
        JButton close = new JButton("Close");
        close.setToolTipText("Close this statistics dialog.");
        close.addActionListener(dialog);
        contentPane.add(close, BorderLayout.SOUTH);
        
        dialog.setVisible(true);
    }
    /**
     Shows a dialog to calculate a value's percentile.
     @see Statistics#percentile(double)
     */
    public void showPercentileDialog()
    {
        JFrame3 dialog = new JFrame3();
        dialog.setSize(800, 150);
        dialog.setTitle("Percentile");
        Container contentPane = dialog.getContentPane();
        
        contentPane.setLayout(new GridLayout(1,2));
        
        JScrollPane scrollPane = new JScrollPane(dialog.getDataTable());
        contentPane.add(scrollPane);
        
        JPanel infoPanel = new JPanel();
        infoPanel.setLayout(new GridLayout(3,1));

        infoPanel.add(new JLabel("Select a cell to see its percentile in the data series:"));
        infoPanel.add(dialog.getOutput());
        
        JButton close = new JButton("Close");
        close.setToolTipText("Close this statistics dialog.");
        close.addActionListener(dialog);
        infoPanel.add(close);
        
        contentPane.add(infoPanel);
                
        dialog.setVisible(true);
    }
    
    //Class for showPValueDialog()
    private class JFrame0 extends JFrame implements java.io.Serializable, ActionListener
    {
        private JTextField textField;
        private JTextArea textArea;
        
        public void actionPerformed(ActionEvent e)
        {
            if(e.getActionCommand().equals("Display"))
            {
                String value = textField.getText();
                if((value!=null)&&(value.length()>0))
                    textArea.setText("The P-value of " + value + " is " + format(stats.pValue(Double.parseDouble(value))) + ".");
            }
            else if(e.getActionCommand().equals("Close"))
                this.dispose();
            else
                System.out.println("Error in user interface.");
        }
    }
    
    //class for showRelativeFrequency()
    private class JFrame1 extends JFrame implements java.io.Serializable, ActionListener
    {
        private JTextField maxField;
        private JTextField minField;
        private JLabel output;
        
        public void actionPerformed(ActionEvent e)
        {
            if(e.getActionCommand().equals("Display"))
            {
                if((maxField.getText()!=null)&&(minField.getText()!=null))
                        output.setText("The relative frequency of values between those limits is " + 
                                format(stats.relativeFrequency(Double.parseDouble(minField.getText()), Double.parseDouble(maxField.getText())))+".");
            }
            else if(e.getActionCommand().equals("Close"))
                this.dispose();
            else
                System.out.println("Error in user interface.");
        }
    }
    
    //class for showOutliers
    private class JFrame2 extends JFrame implements java.io.Serializable, ActionListener
    {
        public void actionPerformed(ActionEvent e)
        {
            if(e.getActionCommand().equals("Close"))
                this.dispose();
            else
                System.out.println("Error in user interface.");
        }
    }
    
    //class for showPercentileDialog()
    private class JFrame3 extends JFrame implements java.io.Serializable, ActionListener, MouseListener, KeyListener
    {
        private JTable data;
        private JTextArea output;
        private int selectedCellIndex;
        private double[] a;
        
        public JFrame3()
        {
            super();
            
            String[] title = {columnTitle};     //Dummy; only used in next line
            data = new JTable(new Object[stats.getSize()][1], title);
            a = stats.getValues();
            for(int i=0; i<a.length; i++)   
                data.setValueAt(Double.toString(a[i]), i, 0);
            data.setCellSelectionEnabled(true);
            data.addMouseListener(this);
            data.addKeyListener(this);

            selectedCellIndex = 0;
            
            output = new JTextArea(1, 20);
            update();
        }
        
        public JTable getDataTable()
        {
            return data;
        }
        
        public JTextArea getOutput()
        {
            return output;
        }
        
        public void actionPerformed(ActionEvent e)
        {
            if(e.getActionCommand().equals("Close"))
                this.dispose();
            else
                System.out.println("Error in user interface.");
        }
        //Check which cell was clicked, and update the output field accordingly.
        public void mouseClicked(MouseEvent e)
        {
            update();              
        }
        
        public void keyReleased(KeyEvent e)
        {
            update();
        }
        
        public void update()
        {
            if(data.getSelectedRow()>=0)
                selectedCellIndex = data.getSelectedRow();
            else        //In case no row is selected.
                selectedCellIndex = 0;
            output.setText("Percentile: " + format(stats.percentile(a[selectedCellIndex])));
        }
        
        //These next few lines just implement the other MouseListener methods.
        public void mouseEntered(MouseEvent e){}
        public void mouseExited(MouseEvent e){}
        public void mousePressed(MouseEvent e){}
        public void mouseReleased(MouseEvent e){}
        //These next few lines just implement the other KeyListener methods.
        public void keyPressed(KeyEvent e){}
        public void keyTyped(KeyEvent e){}
    }
}