package musr.client.gui;
import java.util.Vector;
import javax.swing.*;

public class Statistics implements java.io.Serializable
{
    private double[] stats;       //Holds raw data
    private double mean=0;        //Holds mean of data
    private double adjMean=0;     //Holds mean of data not including outliers
    private double stdDev=0;      //Holds standard deviation of data
    private double adjDev=0;      //Holds standard deviation of data not including outliers
    private double median=0;      //Holds median of data
    private double minimum=0;     //Holds lowest value in data
    private double maximum=0;     //Holds greatest value in data
    private double q1=0;          //Holds the first quartile of the data
    private double q3=0;          //Holds the third quartile of the data
    private double iqr=0;         //Holds the interquartile range of the data
    private double bSkewness = 0;  //Holds Bowley skewness of data
    private double skewness = 0;    //Holds skewness of data
    
    /**
     Creates a new Statistics object.
     This object is based on the objects (assumed to be of type Double) in the Vector newStats. 
     The constructor also sorts the data using a merge sort algorithm.
     @param newStats This Vector is assumed to contain Double values
     */
    public Statistics(Vector newStats)      //Assume it is a vector of Doubles, like a DataColumn
    {
        try //For exception if Vector contains non-Double data.
        {
            int i;
            //Make sure the Vector contains Doubles! If not, throw exception:
            /*for(i=0; i<newStats.size(); i++)
                if(!(newStats.elementAt(i) instanceof Double))
                    throw new Exception("Error: this column contains data which is not of type Double. Ignoring non-Double values.");*/
            //Check Vector for NaN values, which mess things up, and remove them:
            /*for(i=0; i<newStats.size(); i++)
                if(Double.isNaN((Double.parseDouble(((Double)newStats.elementAt(i)).toString()))))
                    newStats.removeElementAt(i);*/
            //In case someone somehow has selected an empty column, this stops exceptions:
            if(newStats.size()==0)
                newStats.addElement(new Double(0));
            stats = new double[newStats.size()];
            for(i=0; i<newStats.size(); i++)
                /*stats[i] = Double.parseDouble(((Double)newStats.elementAt(i)).toString());*/
                stats[i] = Double.parseDouble(newStats.elementAt(i).toString());
            //Sort the array (makes statistics much easier to calculate, so do this first):
            MergeSort sorter = new MergeSort(); 
            this.stats = sorter.sort(this.stats);
        }
        catch(Exception e)      //Takes all the non-Double values out, and initializes this.stats with whatever remains, just as above.
        {
            int i;
            JOptionPane.showMessageDialog(null, e.getMessage());
            Vector tryAgain = new Vector();
            //Fill tryAgain with all the Double values from newStats:
            for(i=0; i<newStats.size(); i++)
                if(newStats.elementAt(i) instanceof Double)
                    tryAgain.addElement(newStats.elementAt(i));
            //Go on with code above, using tryAgain instead of newStats:
            for(i=0; i<tryAgain.size(); i++)
                if(Double.isNaN((Double.parseDouble(((Double)tryAgain.elementAt(i)).toString()))))
                    tryAgain.removeElementAt(i);
            //In case someone somehow has selected an empty column, this stops exceptions:
            if(tryAgain.size()==0)
                tryAgain.addElement(new Double(0));
            stats = new double[tryAgain.size()];
            for(i=0; i<tryAgain.size(); i++)
                stats[i] = Double.parseDouble(((Double)tryAgain.elementAt(i)).toString());
            //Sort the array (makes statistics much easier to calculate, so do this first):
            MergeSort sorter = new MergeSort(); 
            this.stats = sorter.sort(this.stats);
        }
    }  
    
    /**
     @return The mean of the data set.
     */
    public double getMean()
    {
        return mean;
    }
    /**
     @return The mean of the data set, ignoring any outlying values.
     */
    public double getMeanWithoutOutliers()
    {
        return adjMean;
    }
    /**
     @return The standard deviation of the data set.
     */
    public double getStdDev()
    {
        return stdDev;
    }
    /**
     @return The standard deviation of the data set, ignoring any outlying values.
     */
    public double getStdDevWithoutOutliers()
    {
        return this.adjDev;
    }
    /**
     @return The median value of the data set.
     */
    public double getMedian()
    {
        return median;
    }
    /**
     @return The greatest value of the data set.
     */
    public double getMax()
    {
        return maximum;
    }
    /**
      @return The least value of the data set.
     */
    public double getMin()
    {
        return minimum;
    }
    /**
     @return Skewness of the data set.
     */
    public double getSkewness()
    {
        return this.skewness;
    }
    
    /**
     @return Bowley skewness of the data set.
     */
    public double getBowleySkewness()
    {
        return bSkewness;
    }
    /**
     @return The number of items in the data set.
     */
    public int getSize()
    {
        return stats.length;
    }
    /**
     @return A copy of the array of values.
     */
    public double[] getValues()
    {
       double[] result = new double[this.stats.length];
       for(int i=0; i<result.length; i++)
           result[i] = stats[i];
       return result;
    }
    /**
     *Performs calculations to find statistical values.
     This method must be called before calling the accessor methods; otherwise, they return zero.
     */
    public void calculate()
    {
        double sum=0.0;
        //This loop deals with mean.
        for(int i=0; i<this.stats.length; i++)
            sum = sum + stats[i];
        this.mean = sum/(stats.length);
        
        //This loop deals with standard deviation.
        double diffSquared = 0.0;
        for(int i=0; i<this.stats.length; i++)
            diffSquared = diffSquared + Math.pow((stats[i]-mean), 2);
        this.stdDev = Math.sqrt(diffSquared/this.stats.length);
        
        //Maximum and minimum (easy for sorted array); need corrections to prevent errors when reading from very small columns
        if(this.stats.length>1)
            this.maximum = this.stats[this.stats.length-1];
        else
            this.maximum = this.stats[0];
        this.minimum = this.stats[0];
        
        //To find the median:
        this.median = findMedian(this.stats);
        
        //To find the quartiles and interquartile range (which are then used to find Bowley skewness and elsewhere used to find outliers):
        double[] front = new double[this.stats.length/2];
        double[] tail = new double[this.stats.length/2];
        if(stats.length%2==0)
            for(int i=0; i<front.length; i++)
            {
                front[i] = stats[i];
                tail[i] = stats[i + front.length];
            }
        else if(stats.length%2==1)
            for(int i=0; i<front.length; i++)
            {    
                front[i] = stats[i];
                tail[i] = stats[i + front.length + 1];
            }
        this.q1 = findMedian(front);
        this.q3 = findMedian(tail);
        this.iqr = q3-q1;
        if(iqr!=0)      //Prevents div by zero.
        {
            if(stats.length%2==1)       //Odd number of items, so "q2" (middle item) must be included
                this.bSkewness = (q1 - 2*stats[stats.length/2] + q3)/(q3-q1);
            else
                this.bSkewness = (q1+q3)/(q3-q1);
        }
        else
            this.bSkewness = 0.0;
     //To find the skewness:
        double s1 = sum;
        double s2 = 0;
        double s3 = 0;
        for(int i=0; i<this.stats.length; i++)
        {
            s2 = s2 + Math.pow(this.stats[i], 2);
            s3 = s3 + Math.pow(this.stats[i], 3);
        }
        //Broken up into the two k-constants just to space things out:
        double k3 = (2*Math.pow(s1, 3) - 3*stats.length*s1*s2 + Math.pow(stats.length, 2)*s3)/(stats.length*(stats.length-1)*(stats.length-2));
        double k2 = (stats.length*s2-Math.pow(s1, 2))/(stats.length*(stats.length-1)); 
        if((k2!=0)&&(!Double.isNaN(k3))&&(!Double.isNaN(k2)))
            this.skewness = k3/Math.pow(k2, 1.5);
        else
            this.skewness = 0;
        
    //To find the mean without outliers:
        double[] outlierHolder = this.outliers();
        sum = 0;
        for(int i=0; i<stats.length; i++)
            if(!isInArray(stats[i], outlierHolder))
                sum = sum + stats[i];
        this.adjMean = sum/this.stats.length;
        
    //To find the standard deviation without outliers:
        diffSquared = 0.0;
        for(int i=0; i<this.stats.length; i++)
            if(!isInArray(stats[i], outlierHolder))
                diffSquared = diffSquared + Math.pow((stats[i]-adjMean), 2);
        this.adjDev = Math.sqrt(diffSquared/this.stats.length);
    }
    //Helper method for calculate() that finds the median value of a sorted double array. Not exactly complex code.
    private double findMedian(double[] i)
    {
        double result=0;
        if(i.length>=1)
        {
            if(i.length%2==1)  //Odd number of elements, so return middle element
                result = i[i.length/2];
            else              //Even number of elements, so return mean of middle elements
                result = (i[i.length/2] + i[i.length/2-1])/2;
        }
        return result;
    }
    //Helper method to find a double in a double[]
    private boolean isInArray(double value, double[] a)
    {
        for(int i=0; i<a.length; i++)
            if(a[i]==value)
                return true;
        return false;
    }
    /**
     @return P-value (the probability that a value from the data set would be greater than or equal to x).
     */
    public double pValue(double x)      
    {
        int counter = 0;
        while((counter<stats.length)&&(this.stats[counter]<x))
            counter++;
        return (double)(stats.length-counter)/(double)(stats.length);
    }
    /**
      @return The argument's percentile within the data set.
     */
    public double percentile(double x)
    {
        return (1-pValue(x))*100;
    }
    /**
     @return The relative frequency of values between low and high, inclusive.
     */
    public double relativeFrequency(double low, double high)
    {
        int lowerLimit=0;
        while((lowerLimit<stats.length)&&(stats[lowerLimit]<low))
            lowerLimit++;
        int count=0;
        while(((count+lowerLimit)<stats.length)&&(stats[count+lowerLimit]<=high))
            count++;
        return (double)count/(double)stats.length;
    }
    /**
     Finds all data points found to be outliers when compared to the rest of the data set. 
     An outlier is defined as a point which falls more than 1.5 times the interquartile range 
     above the third quartile or below the first quartile.
     @return a Vector holding all outlying values, stored as type Double
     */
    public double[] outliers()
    {
        Vector temp = new Vector();
        for(int i=0; i<stats.length; i++)
        {
            if(((stats[i]<q1)&&((q1-stats[i])>=1.5*iqr))||((stats[i]>q3)&&((stats[i]-q3)>=1.5*iqr)))
                temp.add(new Double(stats[i]));
        }
        //Copy temp into an array:
        double[] outliers = new double[temp.size()];
        for(int i=0; i<temp.size(); i++)
            outliers[i] = Double.parseDouble(((Double)temp.elementAt(i)).toString());
        return outliers;
    }
    
    private class MergeSort     //Sorts the data in the array
    {
        public double[] sort(double[] a)
        {
            int length = a.length;
            if(length<2)
            {}      //Do nothing, and let the method return a.
            else if(length==2)
            {
                if(a[0]>a[1])
                {
                    double temp = a[0];
                    a[0] = a[1];
                    a[1] = temp;
                }
            }
            else if(!isSorted(a))
            {
                double[] front = new double[length/2];
                double[] tail = new double[length-front.length];

                divide(a, front, tail);
                sort(front);
                sort(tail);
                merge(a, front, tail);
            }
            return a;
        }
        
        private void divide(double[] a, double[] front, double[] tail)
        {
            for(int i=0; i<front.length; i++)
                front[i] = a[i];
            for(int i=0; i<tail.length; i++)
                tail[i] = a[front.length + i];
        }
        
        private void merge(double[] a, double[] front, double[] tail)
        {
            int aIndex = 0;
            int frontIndex = 0;
            int tailIndex = 0;
            while((frontIndex<front.length)&&(tailIndex<tail.length))
            {
                if(front[frontIndex]<tail[tailIndex])
                {
                    a[aIndex] = front[frontIndex];
                    aIndex++;
                    frontIndex++;
                }
                else
                {
                    a[aIndex] = tail[tailIndex];
                    aIndex++;
                    tailIndex++;
                }
            }
            while(frontIndex<front.length)
            {
                a[aIndex] = front[frontIndex];
                aIndex++;
                frontIndex++;
            }
            while(tailIndex<tail.length)
            {
                a[aIndex] = tail[tailIndex];
                aIndex++;
                tailIndex++;
            }
        }
        
        private boolean isSorted(double[] a)
        { 
            for(int i=0; i<a.length-1; i++)
                if(a[i]>a[i+1])
                    return false;
            return true;
        }
    }
}